/*
 * This file is part of ccsp.
 * Copyright (C) 2014  Guillaume Quintin.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ccsp.h>

int check_pool(ccsp_pool p,size_t hash_size,size_t interned) {
  size_t cnt,i;
  cnt = 0;
  for( i = 0 ; i < p[0].n ; i++ ) {
    if ( p[0].tab[i] ) {
      struct ccsp_pool_bucket_struct *b;
      for( b = p[0].tab[i] ; b ; b = b->next,cnt++ );
    }
  }
  return !((hash_size == p[0].n) && (cnt == interned));
}

int check_strbuf(ccsp_strbuf buf,const char *s) {
  size_t len;
  if ( buf[0].data == NULL ) return -1;
  len = strlen(buf[0].data);
  if ( len != buf[0].len ) return -1;
  if ( len + 1 >= buf[0].mem ) return -1;
  return strcmp(buf[0].data,s);
}

int check_string(ccsp_string s,int interned,const char *t) {
  if ( s[0].b ) {
    if ( interned == 0 ) return -1;
    return strcmp(s[0].b->data,t);
  }
  else {
    if ( interned != 0 ) return -1;
  }
  return 0;
}

void exit_error2(const char *filename,unsigned long lineno) {
  fprintf(stderr,"\n\n ==> ERROR at %s:%lu\n\n",filename,lineno);
  exit(EXIT_FAILURE);
}

#define exit_error() exit_error2(__FILE__,__LINE__)

void test_ccsp_string_init(void) {
  ccsp_string a;

  ccsp_string_init(a);
  if ( check_string(a,0,NULL) ) exit_error();
  ccsp_string_clear(a);
}

void test_ccsp_pool_init(void) {
  ccsp_pool p;

  ccsp_pool_init(p);
  if ( check_pool(p,1024,0) ) exit_error();
  ccsp_pool_clear(p);
}

void test_ccsp_strbuf_init(void) {
  ccsp_strbuf buf;

  ccsp_strbuf_init(buf);
  if ( check_strbuf(buf,"") ) exit_error();
  ccsp_strbuf_clear(buf);
}

void test_ccsp_strbuf_append(void) {
  ccsp_strbuf buf;

  ccsp_strbuf_init(buf);
  ccsp_strbuf_append(buf,'a');
  ccsp_strbuf_append(buf,'b');
  ccsp_strbuf_append(buf,'c');
  if ( check_strbuf(buf,"abc") ) exit_error();
  ccsp_strbuf_clear(buf);
}

void test_ccsp_intern_strbuf(void) {
  ccsp_strbuf buf;
  ccsp_pool p;
  ccsp_string s;

  ccsp_pool_init(p);
  ccsp_strbuf_init(buf);
  ccsp_string_init(s);
  ccsp_strbuf_append(buf,'a');
  ccsp_strbuf_append(buf,'b');
  ccsp_strbuf_append(buf,'c');
  ccsp_intern_strbuf(s,p,buf);
  if ( check_pool(p,1024,1) ) exit_error();
  if ( check_string(s,1,"abc") ) exit_error();
  ccsp_string_clear(s);
  ccsp_strbuf_clear(buf);
  ccsp_pool_clear(p);
}

void test_ccsp_intern_copy(void) {
  ccsp_pool p;
  ccsp_string s1,s2;
  
  ccsp_pool_init(p);
  ccsp_string_inits(s1,s2,NULL);
  ccsp_intern_cstring(s1,p,"abc");
  ccsp_string_copy(s2,s1);
  if ( check_string(s1,1,"abc") ) exit_error();
  if ( check_string(s2,1,"abc") ) exit_error();
  if ( check_pool(p,1024,1) ) exit_error();
  ccsp_string_clear(s1);
  if ( check_pool(p,1024,1) ) exit_error();
  ccsp_string_clear(s2);
  if ( check_pool(p,1024,0) ) exit_error();
  ccsp_pool_clear(p);
}

void test_ccsp_string_value(void) {
  ccsp_pool p;
  ccsp_string s1,s2;
  
  ccsp_pool_init(p);
  ccsp_string_inits(s1,s2,NULL);
  ccsp_intern_cstring(s1,p,"abc");
  ccsp_string_set_value(s1,(void*)1);
  if ( ccsp_string_get_value(s1) != (void*)1 ) exit_error();
  if ( ccsp_string_get_value(s2) != NULL ) exit_error();
  ccsp_string_clears(s1,s2,NULL);
  ccsp_pool_clear(p);
}

int main(void) {
  test_ccsp_string_init();
  test_ccsp_pool_init();
  test_ccsp_strbuf_append();
  test_ccsp_intern_strbuf();
  test_ccsp_intern_copy();
  test_ccsp_string_value();
  fputs("\n\n ==> OK\n\n",stderr);
  return 0;
}
