/*
 * This file is part of ccsp.
 * Copyright (C) 2014  Guillaume Quintin.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <ccsp.h>

/* just for not writing "struct" */
typedef struct ccsp_pool_bucket_struct ccsp_pool_bucket;

/* function prototypes */
static void ccsp_pool_bucket_delete(ccsp_pool_bucket*);

/********************************************************************
                         memory management
********************************************************************/

static void *ccsp_default_malloc(size_t n) {
  void *p;
  p = malloc(n);
  if ( !p ) {
    fprintf(stderr,"ccsp: cannot malloc %lu bytes\n",
            (unsigned long)n);
    abort();
  }
  return p;
}

static void *ccsp_default_realloc(void *ptr,size_t n) {
  void *p;
  p = realloc(ptr,n);
  if ( !p ) {
    fprintf(stderr,"ccsp: cannot realloc %lu bytes\n",
            (unsigned long)n);
    abort();
  }
  return p;
}

static void ccsp_default_free(void *p) {
  free(p);
}

/* default memory functions */
static void *(*ccsp_malloc)(size_t) = ccsp_default_malloc;
static void *(*ccsp_realloc)(void*,size_t) = ccsp_default_realloc;
static void (*ccsp_free)(void*) = ccsp_default_free;

void ccpp_set_memory_functions(void *(*malloc_func)(size_t),
                               void *(*realloc_func)(void*,size_t),
                               void (*free_func)(void*))
{
  if ( malloc_func == NULL ) malloc_func = ccsp_default_malloc;
  if ( realloc_func == NULL ) realloc_func = ccsp_default_realloc;
  if ( free_func == NULL ) free_func = ccsp_default_free;
  ccsp_malloc = malloc_func;
  ccsp_realloc = realloc_func;
  ccsp_free = free_func;
}

/********************************************************************
                      string buffers stuff
********************************************************************/

void ccsp_strbuf_empty(ccsp_strbuf buf) {
  buf[0].len = 0;
  buf[0].data[0] = 0;
}

void ccsp_strbuf_init2(ccsp_strbuf buf,size_t mem) {
  buf[0].mem = mem;
  buf[0].data = ccsp_malloc(mem * sizeof(char));
  ccsp_strbuf_empty(buf);
}

void ccsp_strbuf_init(ccsp_strbuf buf) {
  ccsp_strbuf_init2(buf,8);
}

void ccsp_strbuf_inits(ccsp_strbuf_ptr buf,...) {
  va_list ap;
  ccsp_strbuf_init(buf);
  va_start(ap,buf);
  for(;;) {
    ccsp_strbuf_ptr ptr;
    ptr = va_arg(ap,ccsp_strbuf_ptr);
    if ( ptr == NULL ) break;
    ccsp_strbuf_init(ptr);
  }
  va_end(ap);
}

void ccsp_strbuf_clear(ccsp_strbuf buf) {
  ccsp_free(buf[0].data);
}

void ccsp_strbuf_clears(ccsp_strbuf_ptr buf,...) {
  va_list ap;
  ccsp_strbuf_clear(buf);
  va_start(ap,buf);
  for(;;) {
    ccsp_strbuf_ptr ptr;
    ptr = va_arg(ap,ccsp_strbuf_ptr);
    if ( ptr == NULL ) break;
    ccsp_strbuf_clear(ptr);
  }
  va_end(ap);
}

static void ccsp_strbuf_grow(ccsp_strbuf buf) {
  buf[0].mem *= 2;
  buf[0].data = ccsp_realloc(buf[0].data,buf[0].mem * sizeof(char));
}

size_t ccsp_strbuf_len(ccsp_strbuf buf) {
  return buf[0].len;
}

void ccsp_strbuf_append(ccsp_strbuf buf,char c) {
  size_t len;
  len = buf[0].len;
  if ( (len + 1) == buf[0].mem ) ccsp_strbuf_grow(buf);
  buf[0].data[len] = c;
  len++;
  buf[0].data[len] = 0;
  buf[0].len = len;
}

char *ccsp_strbuf_get_cstring(ccsp_strbuf buf) {
  return buf[0].data;
}

/********************************************************************
                      string pool stuff
********************************************************************/

static char *strdup2(const char *s) {
  size_t len;
  char *ret;

  len = strlen(s) + 1;
  ret = ccsp_malloc(len * sizeof(char));
  strcpy(ret,s);
  return ret;
}
                      
void ccsp_string_init(ccsp_string s) {
  s[0].b = NULL;
}

void ccsp_string_inits(ccsp_string_ptr s,...) {
  va_list ap;
  ccsp_string_init(s);
  va_start(ap,s);
  for(;;) {
    ccsp_string_ptr ptr;
    ptr = va_arg(ap,ccsp_string_ptr);
    if ( ptr == NULL ) break;
    ccsp_string_init(ptr);
  }
  va_end(ap);
}

void ccsp_string_clear(ccsp_string s) {
  ccsp_pool_bucket *b;
  b = s[0].b;
  if ( b == NULL ) return;
  if ( b->count == 1 ) {
    ccsp_pool_bucket_delete(b);
  }
  else {
    b->count--;
  }
  s[0].b = NULL;
}

void ccsp_string_clears(ccsp_string_ptr s,...) {
  va_list ap;
  ccsp_string_clear(s);
  va_start(ap,s);
  for(;;) {
    ccsp_string_ptr ptr;
    ptr = va_arg(ap,ccsp_string_ptr);
    if ( ptr == NULL ) break;
    ccsp_string_clear(ptr);
  }
  va_end(ap);
}

static void ccsp_string_reset(ccsp_string r) {
  ccsp_string_clear(r);
  ccsp_string_init(r);
}

void ccsp_string_copy(ccsp_string r,ccsp_string s) {
  ccsp_string_reset(r);
  r[0].b = s[0].b;
  if ( r[0].b ) r[0].b->count++;
}
                  
/* This is Dan Berstein's hash function */
static size_t ccsp_default_hash(const char *s,size_t n) {
  size_t ret;
  unsigned char *c;

  ret = 5381;
  for( c = (unsigned char*)s ; c[0] ; c++ ) {
    ret = ret * 33 + ((size_t)c[0]);
  }
  return ret % n;
}

void ccsp_pool_init2(ccsp_pool p,size_t n,
      size_t (*hash_func)(const char*,size_t))
{
  size_t i;

  p[0].n = n;
  p[0].hash = ccsp_default_hash;
  p[0].tab = ccsp_malloc(n * sizeof(ccsp_pool_bucket*));
  for( i = 0 ; i < n ; i++ ) {
    p[0].tab[i] = NULL;
  }
  if ( hash_func == NULL ) {
    p[0].hash = ccsp_default_hash;
  }
  else {
    p[0].hash = hash_func;
  }
}

void ccsp_pool_init(ccsp_pool p) {
  ccsp_pool_init2(p,1024,NULL);
}

void ccsp_pool_inits(ccsp_pool_ptr p,...) {
  va_list ap;
  ccsp_pool_init(p);
  va_start(ap,p);
  for(;;) {
    ccsp_pool_ptr ptr;
    ptr = va_arg(ap,ccsp_pool_ptr);
    if ( ptr == NULL ) break;
    ccsp_pool_init(ptr);
  }
  va_end(ap);
}

void ccsp_pool_clear(ccsp_pool p) {
  size_t i;
  ccsp_pool_bucket *b1,*b2;

  for( i = 0 ; i < p[0].n ; i++ ) {
    for( b1 = p[0].tab[i] ; b1 ; ) {
      b2 = b1->next;
      ccsp_free(b1->data);
      ccsp_free(b1);
      b1 = b2;
    }
  }
  ccsp_free(p[0].tab);
}

void ccsp_pool_clears(ccsp_pool_ptr p,...) {
  va_list ap;
  ccsp_pool_clear(p);
  va_start(ap,p);
  for(;;) {
    ccsp_pool_ptr ptr;
    ptr = va_arg(ap,ccsp_pool_ptr);
    if ( ptr == NULL ) break;
    ccsp_pool_clear(ptr);
  }
  va_end(ap);
}

static void ccsp_pool_bucket_delete(ccsp_pool_bucket *b) {
  if ( b->prev ) {
    b->prev->next = b->next;
  }
  else {
    b->p->tab[b->hash] = b->next;
  }
  if ( b->next ) b->next->prev = b->prev;
  ccsp_free(b->data);
  ccsp_free(b);
}

static void ccsp_intern_cstring2(ccsp_string r,ccsp_pool p,
                                 char *s,int copy)
{
  size_t i;
  ccsp_pool_bucket *b,*prev,*new;
  
  /* first, check if r is interned and ``r == s" */
  if ( r[0].b && r[0].b->p == p && !strcmp(r[0].b->data,s) ) {
      r[0].b->count++;
      return;
  }
  /* reset r */
  ccsp_string_reset(r);

  i = p[0].hash(s,p[0].n);
  if ( p[0].tab[i] == NULL ) {
    /* bucket is empty */
    new = ccsp_malloc(sizeof(ccsp_pool_bucket));
    new->data = ( copy ? strdup2(s) : s );
    new->hash = i;
    new->count = 0;
    new->p = p;
    new->prev = NULL;
    new->next = NULL;
    new->value = NULL;
    p[0].tab[i] = new;
    goto lbl_fill_r;
  }
  /* bucket not empty => check if s is already interned */
  for( b = p[0].tab[i],prev = NULL ; b ; prev = b,b = b->next ) {
    if ( !strcmp(b->data,s) ) {
      /* s is already interned */
      new = b;
      goto lbl_fill_r;
    }
  }
  /* string is not interned yet */
  new = ccsp_malloc(sizeof(ccsp_pool_bucket));
  new->data = ( copy ? strdup2(s) : s );
  new->hash = i;
  new->count = 0;
  new->p = p;
  new->prev = prev;
  new->next = NULL;
  new->value = NULL;
  prev->next = new;

lbl_fill_r:
  r[0].b = new;
  new->count++;
}

void ccsp_intern_strbuf(ccsp_string r,ccsp_pool p,ccsp_strbuf buf) {
  ccsp_intern_cstring2(r,p,ccsp_strbuf_get_cstring(buf),1);
}

void ccsp_intern_cstring(ccsp_string r,ccsp_pool p,const char *s) {
  ccsp_intern_cstring2(r,p,(char*)s,1);
}

void ccsp_intern_cstring_alias(ccsp_string r,ccsp_pool p,char *s) {
  ccsp_intern_cstring2(r,p,s,0);
}

void ccsp_string_set_value(ccsp_string s,void *value) {
  if ( !s[0].b ) return;
  s[0].b->value = value;
}

void *ccsp_string_get_value(ccsp_string s) {
  if ( !s[0].b ) return NULL;
  return s[0].b->value;
}

char *ccsp_string_get_cstring(ccsp_string s) {
  return ( s[0].b ? s[0].b->data : NULL );
}

